/* Field.java - Copyright (C) 2000 Ilpo Jrvinen
 *
 * TiRa harjoitusty K2000
 */


import	AWTaddon.*;
import	java.io.*;
import	java.awt.*;
import	java.awt.event.*;
import	java.util.Enumeration;


/*******************************************************************
 * This class is used for drawing the map and everything that is 
 * related to it into the screen.
 *
 * @author	Ilpo Jrvinen <ijjarvin@cs.helsinki.fi>
 * @version 0.97f
 */
public class Field extends Canvas implements MouseListener {

	/** Size of map square in pixels */
	public static final int		SPACING = 16;
	
	private Robot				parent = null;						// save & load & generate need this.
	private Map 				map = null;
	private Path				path = null;
	private Point				startPoint = null;					// temporary storage

	private String				lastDir = null;						// save & load last working dir

	Dimension preferredSize = null;									// docs say these are required
	Dimension minimumSize = new Dimension(SPACING*2, SPACING*2);	// but I don't know how they work
	
	
	/*************************************************************
	 * Constructs empty (mapless) field.
	 *
	 * @param	parent	parent window.
	 */
	public Field( Robot parent )
	{
		this.parent = parent;
		preferredSize = minimumSize;
		addMouseListener( this );
	}
	

	/************************************************************
	 * Sets new map.
	 *
	 * @param 	map		the map
	 */
	public void setMap( Map map )
	{
		path = null;							// clear info that is map
		startPoint = null;						// depentant
		
		this.map = map;							// set new map, synchronize?
		if (map != null)
			preferredSize = new Dimension( SPACING * map.getXSize(), SPACING * map.getYSize() );
		else
			preferredSize = minimumSize;
		repaint();								// redraw...
	}
	
	
	/***********************************************************
	 * Displays the map, path & startpoint if valid. 
	 */
	public void paint( Graphics g ) 
	{
		if (map == null)
			return;
		
		g.setColor( Color.gray );
		drawGrid( g );

 		g.setColor( Color.black );
		drawObstacles( g );

		drawPath( g );				// Color is path/no path dependant.		


		g.setColor( Color.red );
		drawCircle( g, startPoint );
	}


	/***************************************************************
	 * This method draws the "grid" around each square.
	 */
	private void drawGrid( Graphics g )
	{
		for (int i = 0; i <= SPACING * map.getYSize(); i += SPACING) {
			g.drawLine( 0, i, SPACING * map.getXSize(), i );
		}
		for (int i = 0; i <= SPACING * map.getXSize(); i += SPACING) {
			g.drawLine( i, 0, i, SPACING * map.getYSize() ); 		
		}
	}


	/****************************************************************
	 * This method draws the obstacles on the current map.
	 */	
	private void drawObstacles( Graphics g )
	{
		for (int i = 0; i < map.getXSize(); i++) {
			for (int j = 0; j < map.getYSize(); j++) {
				if (map.isImpassable( i, j ))
					g.fillRect( i * SPACING + 1, j * SPACING + 1, SPACING - 1, SPACING - 1 );
			}
		}
	}


	/*************************************************************************
	 * a bit darker than normal green which displays poorly on most monitors.
	 */
	private static final Color pathColor = new Color( 0, 150, 0 );		
	
	
	/************************************************************************
	 * This method draws the path.
	 */
	private void drawPath( Graphics g )
	{			
		if (path != null) {
			g.setColor( Color.white );
			Point start = (Point)path.endPoint(false);
			Point end = (Point)path.endPoint(true);
			drawSquare( g, start );
			drawSquare( g, end );

			if (!path.hasPath()) {
				g.setColor( Color.red );
				Draw.drawBrokenLine( g, start.getX() * SPACING + SPACING / 2, 
									 start.getY() * SPACING + SPACING / 2,
									 end.getX() * SPACING + SPACING / 2, 
									 end.getY() * SPACING + SPACING / 2, 4 );
				
			} else {
				g.setColor( pathColor );
				Enumeration temp = path.elements();
				Point last = (Point)temp.nextElement();
				while( temp.hasMoreElements() ) {
					Point next = (Point)temp.nextElement();
					g.drawLine( last.getX() * SPACING + SPACING / 2, 
								last.getY() * SPACING + SPACING / 2, 
								next.getX() * SPACING + SPACING / 2, 
								next.getY() * SPACING + SPACING / 2 );
					last = next;
				}
			}
		}
	}
	

	/**********************************************************************
	 * This method draws a square to specified Point p on map.
	 * The surrounding grid is left untouched.
	 */	
	private void drawSquare( Graphics g, Point p )
	{
		if (p == null)
			return;								// shouldn't ever happend
		g.fillRect( p.getX() * SPACING + 1, 
					p.getY() * SPACING + 1,
					SPACING - 1, 
					SPACING - 1 );		
	}


	/*********************************************************************
	 * This method draws a circle inside of specified Point p on map.
	 */
	private void drawCircle( Graphics g, Point p )
	{
		if (p == null)
			return;								// shouldn't ever happend
		g.fillOval( p.getX() * SPACING + 1,
					p.getY() * SPACING + 1, 
					SPACING - 2, 
					SPACING - 2);
	}
	


	/*
	 * Docs say: required. (scrollpane functions)
	 * The exact reason is still a mystery to me.
	 */
	public Dimension getMinimumSize() { return minimumSize; }                    
	public Dimension getPreferredSize() { return preferredSize; }


	/********************************************************************
	 * This event handler handles endpoint selection.
	 *
	 * @param	e	MouseEvent from AWT.
	 */
	public void mouseReleased(MouseEvent e) 
	{
		int x = e.getX() / SPACING;
		int y = e.getY() / SPACING;

		if (map == null || x >= map.getXSize() || y >= map.getYSize())
			return;								// catch exception trouble

		if (startPoint == null) {
			startPoint = new Point( x, y );		// select first point
		} else {
			Point end = new Point( x, y );		// create second point
			if (!end.equals(startPoint)) {		// if points differ do A*
				path = map.findRoute( startPoint, end );
				if (path.hasPath()) {
					double l = path.getLength();
					l = Math.round( l * 100 ) / 100.0;		// 2 decimal places
					parent.message( "Path length " + l );
				} else {
					parent.message( "No legal path" );
				}
			}
			startPoint = null;					// ...deselect the first point
		}
		repaint();								// redraw the path
	}
	
	
	/*
	 * Flicker avoidance (dblbuffer)
	 */
	public boolean isDoubleBuffered()			{ return true; }	// AWT should know about it...
	public void update( Graphics g )
	{
		if (map == null)											// Exception trap.
			return;
		Image im = createImage( (map.getXSize() + 1) * SPACING,		// create temporary drawing area
								(map.getYSize() + 1) * SPACING );	// wastes some resources, I DON'T CARE!
		Graphics gr = im.getGraphics();
		paint( gr );												// ...and draw it...
		g.drawImage( im, 0, 0, Color.black, null );					// CHECKME??? What a heck is ImageProducer? (last parameter)
		gr.dispose();												// free resources, gc would do, but it takes some time...
		im.flush();													// ...free more resources...

		/*
		 * Java hack, to fix mem release problems on win JDK+AWT
		 * The im.flush() alone won't do the job and eventually leads to
		 * OutOfMemoryError.
		 * The Linux/Solaris JDKs work ok.
		 */
		// System.gc();
		/*
		 * Hack disabled.
		 */
	}
	
	/***************************************************************
	 * This method is the GUI driver for random map generator.
	 */
	public void generate()
	{
		GenerateDialog g = new GenerateDialog( parent );
		g.setVisible( true );
		if (!g.getOk())	
			return;					// cancelled.

		setMap( new Map( g.getXValue(), g.getYValue(), 2 ) );
		parent.message("");
	}
	
	
	/****************************************************************
	 * This method is the GUI driver for map loader.
	 */
	public void load()
	{
		FileDialog f = new FileDialog( parent, "Load map", FileDialog.LOAD );
		dirChange( f );
		f.setVisible( true );
		dirSave( f );
		if (f.getFile() == null)
			return;						// cancelled.
		
		try {
			setMap( new Map( f.getDirectory() + '/' + f.getFile() ) );
			parent.message( "Map loaded" );
			return;
		} catch( OutOfMemoryError ex ) {
			parent.message( "Map file was invalid" );
		} catch( Exception ex ) {
			parent.message( "Map file was invalid" );
		}
		setMap( null );	
	}


	/****************************************************************
	 * This method is the GUI driver for saving the map.
	 */
	public void save()
	{
		if (map == null)
			return;						// no map in program.
		FileDialog f = new FileDialog( parent, "Save map", FileDialog.SAVE );
		dirChange( f );
		f.setVisible( true );
		dirSave( f );
		if (f.getFile() == null)
			return;						// cancelled.
			
		if (map.saveMap( f.getDirectory() + '/' + f.getFile() )) {
			parent.message("Saved ok");
		} else {
			parent.message("Save failed");
		}
	}


	/**************************************************************
	 * This method creates the map directory if necessary, but
	 * it's done only at the first time. Known gotcha, lastDir is
	 * removed between fileops.
	 *
	 * @param f Selection FileDialog.
	 */
	private void dirChange( FileDialog f )
	{
		if (lastDir == null) {
			File dir = new File( "map" );
			if (!dir.exists()) {
//				System.out.println( "\nCreating map directory: " + dir.getAbsolutePath() );
				dir.mkdir();
			}
			lastDir = dir.getAbsolutePath();
		}
		f.setDirectory( lastDir );
	}
	
	/************************************************************
	 * This method saves the last directory used by load/save op
	 *
	 * @param f Selection FileDialog.
	 */
	private void dirSave( FileDialog f )
	{
		lastDir = f.getDirectory();
	}

	

	/*
	 * Dummy interface methods.
	 */
	public void mouseClicked(MouseEvent e) {}
	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}
	public void mousePressed(MouseEvent e) {}

	
} // eoc
