/* Heap.java - Copyright (c) 2000 Ilpo Jrvinen
 *
 * TiRa harjoitusty K2000
 */

/*
 * 09.05.2000	Resize method added, it's dynamic now... :)
 */


/**********************************************************************
 * This class is heap implementation. The heap itself is stored into
 * array heap, in which index zero position is left empty and the "root"
 * will be at position 0 
 *
 * @author 	Ilpo Jrvinen <ijjarvin@cs.helsinki.fi>
 * @version	1.01f
 */
 
import java.util.*;					// NoSuchElement

public class Heap {

	public static int 				DEFAULTCAPACITY = 16;

	private Comparable[] 			heap = null;
	private int 					size = 0;
	private int						resizeCount = 0;			// benchmarking only


	/****************************************************************
	 * Constructs new heap with default capacity (16).
	 */
	public Heap() 
	{
		this( DEFAULTCAPACITY );
	}

	/****************************************************************
	 * Constructs new heap with specified size.
	 *
	 * @param	capacity	the initial capacity of heap.
	 */
	public Heap( int capacity )
	{
		if (capacity < 1) 
			capacity = 1;
		heap = new Comparable[capacity+1];			// index 0 is empty
	}
	
	
	/***********************************************************
	 * This method doubles the size of the heap.
	 * It's called from insert method when heap is out of space.
	 */
	private void resize()
	{
		Comparable temp[] = new Comparable[ heap.length * 2 ];
		System.arraycopy( heap, 0, temp, 0, heap.length );
		heap = temp;
		resizeCount++;
	}

	/** DEBUG ONLY */
	public int	getResizeCount()	{ return resizeCount; }		// benchmarking only


	/*********************************************************
	 * This method inserts c to the heap. Technicue used for finding correct 
	 * position to c is known as percolate up...
	 *
	 * @param	c	element to be inserted to queue.
	 */
	public void insert(Comparable c) 
	{
		if (isFull())						// see if the whole space has been used up...
			resize();						// make it last forever

		int hole = ++size;
		int	parent = hole >> 1;				// / 2;

		while (hole > 1 && c.compareTo(heap[parent]) < 0) {
			heap[hole] = heap[parent];		// perc up
			hole = parent;
			parent >>= 1;					// / 2;
		}

		heap[hole] = c;
	}


	/******************************************************************
	 * This method removes the root of the heap. The removed element is 
	 * quaranteed to be the smallest element in the heap (heap property).
	 *
	 * @return	The smallest element in heap.
	 *
	 * @throw	NoSuchElementException	if heap is empty. 
	 */
	public Comparable deleteMin() 
	{
		if (isEmpty())
			throw new NoSuchElementException();

		Comparable smallest = heap[1];		// save the smallest...
		heap[1] = heap[size];       		// prepare for perc down
		heap[size--] = null;				// clear trash

		int parent = 1, child = 2; 			// ...start at root

		while (child <= size) {  			// ...still has child(ren)
			/*
			 * Select right child if it is smaller than left (and exists)...
			 */
			if (child != size && heap[child].compareTo(heap[child+1]) > 0)
				++child;
 
			if (heap[parent].compareTo(heap[child]) > 0) {
				Comparable temp = heap[parent];		// perc down.
				heap[parent] = heap[child];
				heap[child] = temp;
			} else {
				break;						// the position is now ok.
			}
			
			parent = child;
			child = parent << 1;			// * 2
		}

		return smallest;
	}

	/*
	 * Anything below is stolen from AW's Keko.java (TiRa K2000)
	 */
	

	/******************************************************************
	 * @return	true if heap need to be resized for insertion.
	 */
	private boolean isFull()		{ return size == heap.length - 1; }
	
	
	/******************************************************************
	 * @return	true if the are no elements in this heap.
	 */
	public boolean isEmpty()	{ return size == 0; }

	
	
	/*
	 * Anything below this is for debugging only.
	 */
	/** DEBUG ONLY */
	public String toString() {
		return merkkiEsitys(1, "");
	}
	
	// tmn tulostusasun ideasta kiitos kurssin opiskelijalle
	// Markku J Strandmanille
	/** DEBUG ONLY */
	private String merkkiEsitys(int juuri, String sisennys) {
		if (juuri>size)
			return "";
		return 
			merkkiEsitys(2*juuri+1, sisennys + "     |") +
			"\n" + sisennys + "--->" + heap[juuri]        +
			merkkiEsitys(2*juuri, sisennys + "     |");
	}

/* Resize debugging. Add / in the beginning of this line if you want to debug
	protected void finalize() throws Throwable
	{
		super.finalize();
		System.out.println( "Before finalization this Heap has been resized " +
							getResizeCount() + " times" );
	}
// End of comment */
}
