/* Map.java - Copyright (c) 2000 Ilpo Jrvinen
 *
 * TiRa harjoitusty K2000
 */

import java.io.*;


/*************************************************************************
 * This class contains the map data. The map itself is stored in an array
 * of int, one bit per square. 
 * The routefinder algorithm is also included inside this class because
 * it uses knowledge about structure of the map instead of doing just generic
 * net search.
 *
 * @author	Ilpo Jrvinen <ijjarvin@cs.helsinki.fi>
 * @version	0.99f
 */
 
public class Map {

	/*****************************************************************
	 * Find good value to minimize queue resizes during A* operation.
	 * This is still a lot, I have got 1 resize when nopath from corner
	 * to corner. Also some very complecated maps can cause 1 resize.
	 * 2 is overkill.
	 */
	public	static final int		QUEUEMULTIPLIER = 1;
	
	private	int						map[] = null;
	private int 					xSize = -1;
	private int						ySize = -1;


	/*******************************************************************
	 * This method allocates the array for map, every square is represented 
	 * with a single bit, which is 1 if corresponding square is impassable.
	 *
	 * @param	x	horizontal size of the map.
	 * @param	y	vertical size of the map.
	 *
	 * @return	the allocated array.
	 *
	 * @throw	OutOfMemoryError	if request was non-sense.
	 */
	private int[] allocMap( int x, int y )
	{
		long size = (long)x * (long)y;
		if (x < 0 || y < 0 || size >= Integer.MAX_VALUE)
			throw new OutOfMemoryError();				// braindmg attempt, load can cause this. (Size bigxbig)
		return new int[ ((int)size + 31) / 32 ];
	}


	/*******************************************************************
	 * This constructor creates map w/o any obstacles.
	 *
	 * @param	x	horizontal size of the map.
	 * @param	y	vertical size of the map.
	 *
	 * @throw	OutOfMemoryError	if request was non-sense.
	 */
	public Map( int x, int y )
	{
		xSize = x + 2;									// borders need space too
		ySize = y + 2;
		
		map = allocMap( xSize, ySize );
		addBorder();									// do the border around the map.
	}



	/******************************************************************
	 * Constructs a map with some (mostly) random stuff in it.
	 *
	 * @param	x	horizontal size of the map.
	 * @param	y	vertical size of the map.
	 * @param	d	divisor for randomizer
	 *
	 * @throw	OutOfMemoryError	if request was non-sense.
	 */
	public Map( int x, int y, int div ) 
	{
		this( x, y );
		generate( div );
	}
	
	/******************************************************************
	 * This constructor reads map from file.
	 *
	 * @param	filename	name of the mapfile.
	 *
	 * @throw	Exception			when format is non-sense.
	 * @throw	OutOfMemoryError	when size is too large or negative
	 */
	public Map( String filename ) throws Exception
	{		
		try {
			BufferedReader in = new BufferedReader( 
									new FileReader( 
										new File( filename ) ) );
			if (readSize( in )) {
				map = allocMap( xSize, ySize );
				addBorder();
		
				readData( in );
				return;
			}
			// Map file was not valid, fall to throw
		} catch( Exception e ) { 
//			System.out.println( e );							// Debug
		}
		throw new Exception( "map file corrupted" );		// Exception is the only option!
	}
	

	/***********************************************************************
	 * This method reads & parses the "size nxn" line, returns true if line
	 * was found. Empty lines in the front of it are skipped.
	 *
	 * @param	in	Reader for datastream.
	 *
	 * @return	true if size line was found.
	 */	
	private boolean readSize( BufferedReader in ) throws IOException
	{
		String s;

		while ((s = in.readLine()) != null) {		// skip empty lines
			s = s.trim();
			if (s.length() > 0)
				break;
		}
		if (!s.toLowerCase().startsWith( "size" ))	// check if valid?
			return false;

		/*
		 * cut "size"..., and convert capital X (if has one?) to x
		 * ...then read x & y
		 */
		s = s.substring( s.indexOf( ' ' ) + 1 ).toLowerCase().trim();
		xSize = Integer.parseInt( s.substring(0, s.indexOf( 'x' ) ) ) + 2;
		ySize = Integer.parseInt( s.substring(s.indexOf( 'x' ) + 1 ) ) + 2;
	
		return true;
	}


	/*********************************************************************
	 * This method reads the "picture" of the map from file, the borders
	 * are not read but skipped.
	 *
	 * @param	in	Reader for datastream.
	 */
	public void readData( BufferedReader in ) throws IOException
	{
		String s;
		int y = 1;
		while ((s = in.readLine()) != null) {
			if (y > ySize)
				break;
			for (int x = 1; x < xSize-1; ) {
				if (s.length() == 0)					// this line is out of data?
					break;
				if (s.charAt(0) == '\t')				// tab are 8 chars wide
					x += 8;								// tab is passable * 8
				else if (s.charAt(0) == ' ')			// space is passable * 1
					x++;
				else
					setImpassable( x++, y );			// anything else, is impassable
				s = s.substring( 1 );
			}
			y++;
		}
	}


	/**************************************************************************
	 * This method saves the map to file, the borders are not saved (not shown
	 * at all)
	 *
	 * @return	true if save was successfully completed.
	 */	
	public boolean saveMap( String filename )
	{
		try {
			BufferedWriter out = new BufferedWriter(
									new FileWriter(
										new File( filename ) ) );

			String s = "Size " + getXSize() + "x" + getYSize() + "\n";
			out.write( s, 0, s.length() );			// save size

			for (int j = 0; j < getYSize(); j++) {	// save square data
				for (int i = 0; i < getXSize(); i++) {
					out.write( isImpassable( i, j ) ? '*' : ' ' );
				}
				out.write('\n');
			}
			
			out.flush();							// force commit!
			out.close();
			return true;
			
		} catch( IOException e ) {
			return false;							// ...failed... :(
		}
	}


	/********************************************************************
	 * This method returns the size of the map. The caller doesn't know 
	 * anything about the border. 
	 *
	 * @return	horizontal size of the map w/o hidden border.
	 */
	public int getXSize()				{ return xSize - 2; }	
	
	
	/********************************************************************
	 * This method returns the size of the map. The caller doesn't know 
	 * anything about the border. 
	 *
	 * @return	vertical size of the map w/o hidden border.
	 */
	public int getYSize()				{ return ySize - 2; }



	/********************************************************************
	 * This method generates some lines to map...
	 *
	 * @param	div		divisor...
	 */
	private void generate( int div )
	{
		int c, s, e;						// const, start, end
		int lines = (int)(((xSize + ySize) / div) * Math.random()) + 1;
		
		while (lines-- > 0) {
			if (Math.random() < 0.5) {		// horiz. line
				c = (int)(ySize * Math.random());
				s = (int)(xSize * Math.random());
				e = (int)(xSize * Math.random());
				for (int i = Math.min(s, e); i < Math.max(s, e); i++)
					setImpassable( i, c );
			} else {						// vert. line
				c = (int)(xSize * Math.random());
				s = (int)(ySize * Math.random());
				e = (int)(ySize * Math.random());
				for (int i = Math.min(s, e); i < Math.max(s, e); i++)
					setImpassable( c, i );				
			}
		}
	}

	/********************************************************************
	 * This method checks if (x, y) is not passable.
	 * Borders are ignored! (=0,0 is not on the border!)
	 *
	 * @param	x	Point x co-ordinate	
	 * @param	y	Point y co-ordinate
	 *
	 * @return true if square has an obstacle in it.
	 */
	/*
	 * This method is not called internally due conflicting x,y definations.
	 */
	public boolean isImpassable( int x, int y )
	{
		x++; y++;										// skip borders
		if (x <= 0 || x >= xSize - 1 || y <= 0 || y >= ySize - 1)
			return true;
		int index = y * xSize + x;
		return ((map[index/32] >> (index % 32)) & 0x1) != 0;
	}


	/********************************************************************
	 * This method allows easier access to map when point (x,y) is set
	 * impassable.
	 *
	 * @param	x	Point x coordinate	
	 * @param	y	Point y coordinate
	 */	
	private	void setImpassable( int x, int y )
	{
		int index = y * xSize + x;
		map[ index / 32 ] |= 1 << (index % 32);
	}

	
	/*******************************************************************
	 * This method creates the borders around the area.
	 * The border is never shown outside!
	 */
	private void addBorder()
	{
		for (int i = 0; i < xSize; i++) {
			setImpassable(i, 0);
			setImpassable(i, ySize-1);
		}
		for (int i = 0; i < ySize; i++) {
			setImpassable(0, i);
			setImpassable(xSize-1, i);
		}
	}


	
/************************************************************************
 
				The pathfinder! (A*)
 
 ************************************************************************/

	/********************************************************************
	 * This method searches the Path from start to end. If there aren't any
	 * valid paths, the returned Path will contain only start and endpoints.
	 *
	 * @param	start	The startpoint of the route.
	 * @param	end		The endpoint of the route.
	 *
	 * @return	The Path that was found.
	 */
	public Path findRoute( Point start, Point end )
	{
		// Points cannot be used because the x and y won't match.
		return findRoute( start.getX(), start.getY(), end.getX(), end.getY() );
	}


	/************************************************************************
	 * This method is the actual A* implementation when combined with method
	 * spawnOpen. findRoute initializes the data structures, then starts
	 * deQueueing nodes, one-by-one, calling spawnOpen for each deQueued node.
	 * This method is thread-safe.
	 *
	 * @return The constructed Path.
	 */
 	private Path findRoute( int x1, int y1, int x2, int y2 )
	{
		/* 
		 * prepare data structures...
		 */
		Point destination = new Point( x2 + 1, y2 + 1 );	// This is what we looking for...
		Node visited[] = new Node[ xSize * ySize ];
		Heap queue = new Heap( (xSize + ySize) * QUEUEMULTIPLIER );
		
		/*
		 * Insert the seed only if path can exist...
		 * if start or end loc is impassable, the queue is left empty!
		 * -> correct "no path" return during first iteration.
		 */
		if (!(isImpassable( x1, y1 ) || isImpassable( x2, y2 )))
			queue.insert( new Node( null, x1 + 1, y1 + 1, 0, 0 ) );		// startpoint node
		
		do {
			if (queue.isEmpty())							// no more open squares?
				break;
				
			Node temp = (Node)queue.deleteMin();			// get the "assumed best"
		
			if (destination.equals( temp ))
				return buildPath( temp );					// reached!

			if (!temp.closed) {								// see NOTE 1 below.
				spawnOpen( temp, queue, visited, destination );
				temp.closed = true;
			}
		} while (true);
		
		Path ret = new Path( -1.0 );				// build no path instance of Path.
		ret.addPoint( new Point( x1, y1 ) );		// ...add endpoints
		ret.addPoint( new Point( x2, y2 ) );
		return ret;
	}
	/* 
	 * NOTE 1:
	 *
	 * This check may seem useless at first, but it's not because
	 * decreaseKey is not very trivial to do in java with Comparables.
	 * Check NOTE 2 below. When quicker path to _open_ Node (it has to be 
	 * still open because when Node is closed, the distance from startpoint
	 * to Node is _optimal_, guaranteed by A* algorithm with underestimating
	 * heuretics) is found, instead of doing decreaseKey(), a new Node with 
	 * equal Point, is insert to priority queue. This causes dequeuing of 
	 * Node of Point to occur twice (with different priorities), So the later
	 * one is  simply ignored, because the first one is the _optimal_ route 
	 * from starting point.
	 */				
	
	
	/*
	 * spawnOpen will use these constants.
	 */
	private static final double	SQRT2MINUS1 = Math.sqrt(2) - 1.0;
	private static final double	ONE			= 1.0;
	private static final double	SQRT2		= Math.sqrt(2);

	/* 
	 * Magic additions, 		0	1	2
	 * these save one fetch		7	n	3
	 * this coord = last+delta	6	5	4
	 */
	private static final int	xDelta[]	= { -1, 1, 1, 0, 0, -1, -1, 0 };
	private static final int	yDelta[]	= { -1, 0, 0, 1, 1, 0, 0, -1 };
	private static final double	cost[]		= { SQRT2, ONE, SQRT2, ONE, SQRT2,
												ONE, SQRT2, ONE, SQRT2, ONE };

	/***********************************************************************
	 * This method creates all open nodes surrounding node n and inserts them
	 * to "open list" if move is possible and can actually be on the fastest
	 * route. The "open list" is Heap queue.
	 * 
	 * @param	n	Node whose surrondings are spawn.
	 */
	private void spawnOpen( Node n, Heap queue, Node visited[], Point destination )
	{
		int x = n.getX(), y = n.getY();
				
		for (int i = 0; i < xDelta.length; i++) {
			x += xDelta[i];								// add magic, ?Delta is relative
			y += yDelta[i];								// to the last (x,y), not to node n!
			
			if (!canMove( n.getX(), n.getY(), x, y))	// check validity of the move.
				continue;
			/*
			 * ...First check if we have already been there?
			 * This _can_ and _will_ happend when steps get updated and
			 * node is still "open".
			 */			
			Node temp = visited[y * xSize + x];
			if (temp == null) {							// ...no, create new Node.
				temp = new Node( n, x, y, n.steps + cost[i], heuretics( destination, x, y ) );
				visited[y * xSize + x] = temp;
				queue.insert(temp);

			} else {									// ...yes
				if (temp.steps <= n.steps + cost[i])
					continue;
				temp.steps = n.steps + cost[i];			// ...this route was quicker, fix steps
				temp.parent = n;						// ...and backtracking info
				
				queue.insert(temp);						// see NOTE 2 below.
			}
		}
	} // eom
	
	/* 
	 * NOTE 2:
	 *
	 * Do decreaseKey here instead of insert!
	 * How decreaseKey works with Comparables? They SUCK big time!
	 * TRYME?!? Implement interface Priorizable implements Comparable
	 * but that doesn't solve the double vs int incompatibility. :(
	 * PriorizableDouble? No way, it would just make things messy
	 * and complicated...
	 */


	/***************************************************************
	 * This builds a path from Node seed, using backtracking info
	 * (parent) in nodes. The backtracking starts from Node seed and
	 * ends when parent is null. If parent chain has loops it might
	 * end up into infinite loop, but A* guarantees there won't be
	 * any loops.
	 *
	 * @param	seed	Node from which the backtracking will start.
	 *
	 * @return	Constructed Path.
	 */
	private Path buildPath( Node seed )
	{
		Path ret = new Path( seed.total() );
		// create "the path", this really exists
		while (seed != null) {
			ret.addPoint( new Point( seed.getX()-1, seed.getY()-1 ) );
			seed = seed.parent;						// backtrack
		}
		return ret;		
	}

	/******************************************************************
	 * This method checks if move from (x, y) to (dx, dy) is valid.
	 *
	 * @param	x	Point 1 x co-ordinate	
	 * @param	y	Point 1 y co-ordinate
	 *
	 * @param	x	Point 2 x co-ordinate	
	 * @param	y	Point 2 y co-ordinate
	 *
	 * @return	true if move is valid.
	 */
	private boolean canMove( int x, int y, int dx, int dy )
	{
		/*
		 * Check that destination is clear
		 */
		int index = dy * xSize + dx;
		if (((map[index/32] >> (index % 32)) & 0x1) != 0)
			return false;
		/*
		 * check for corners if diagonal move!
		 */
		if (dx != x && dy != y) {
			index = y * xSize + dx;
			if (((map[index/32] >> (index % 32)) & 0x1) != 0)
				return false;
			index = dy * xSize + x;
			if (((map[index/32] >> (index % 32)) & 0x1) != 0)
				return false;			
		}
		return true;
	}


	/***************************************************************
	 * Estimate the optimal distance between x,y and destination.
	 * The guess is optimitic calculation of minimum distance from
	 * x,y and destination, if there is no obstacles on the way the
	 * guess will be correct, if there are some obstacles, it's still
	 * not completely wrong to guess like it does, because any guess
	 * that _overestimates_, even slightly can result an sub-optimal
	 * performance in the A*-algorithm.
	 *
	 * @param	destination	Point where we want to end up.
	 * @param	x	Point x co-ordinate	
	 * @param	y	Point y co-ordinate
	 *
	 * @return	An guess of distance from x,y to destination.
	 */
	private double heuretics( Point destination, int x, int y )
	{
		int dx = Math.abs(destination.getX() - x);
		int dy = Math.abs(destination.getY() - y);
		return Math.max( dx, dy ) + Math.min( dx, dy ) * SQRT2MINUS1;
	}

	
}

/**************************************************************************
 * Class Node is a "simple" storage class, nothing less, a bit more though
 * 'cause it implements Comparable.
 *
 * @author	Ilpo Jrvinen <ijjarvin@cs.helsinki.fi>
 * @version	0.95f
 */
class Node extends Point implements Comparable {

	private static final double	ZERO			= 0.0;

	protected Node 		parent;
	protected double	steps;
	protected double	heuretics;
	protected boolean	closed;


	/**********************************************************************
	 * This constructor creates new Node, with all necessary attributes.
	 *
	 * @param	par		parent of this node.
	 * @param	x		x coordinate of this node.
	 * @param	y		y coordinate of this node.
	 * @param	cost	a travelling cost from start to this node.
	 * @param	heuretics	an estimated cost for travel from here to end.
	 */
	protected Node( Node par, int x, int y, double cost, double heuretics )
	{
		super( x, y );
		parent = par;
		closed = false;
		this.steps = cost;
		this.heuretics = heuretics;
	}
	

	/*************************************************
	 * @return	estimated togo plus steps so far.
	 */	
	protected double total()
	{
		return steps + heuretics;
	}


	/***********************************************************************
	 * This method compares two Nodes with each other.
	 * The equality defination is not "== equality" because round-off errors
	 * can cause flawed results.
	 * a and b will be considered as equal when following condition applies:
	 * 
	 *   |a.total() - b.total()| < SMALLCONSTANT
	 */	
	private static final double	SMALLCONSTANT	= 0.001;

	public int compareTo( Object o )
	{
		double t = total(), ot = ((Node)o).total();
		if (t - ot < SMALLCONSTANT)
			return -1;
		else if (t - ot > SMALLCONSTANT)		
			return 1;
		else
			return 0;
	}
}
